from neuro_helper.entity import Space


def task_colors(space: Space, with_rest=False):
    if space in [Space.K59_CORTEX, Space.K59]:
        colors = ["#f032e6", "#aaffc3"]
        if with_rest:
            return ["#B89B49"] + task_colors
        else:
            return colors

    raise ValueError(f"no task color is defined for {space}")




from neuro_helper.entity import Space


def task_order(space: Space, with_rest=True):
    if space in [Space.K59, Space.K59_CORTEX]:
        if with_rest:
            out = ["REST", ]
        else:
            out = []
        return out + ["MOVIE", "RET"]
    elif space in [Space.K32, Space.K32_CORTEX]:
        return ["REST", ]

    raise ValueError(f"task order for {space} is not defined")



import re
from scipy.io import loadmat
import numpy as np
from neuro_helper.entity import TemplateName
from neuro_helper.storage import LocalStorage, StorageFile, ANYTHING, RCloneStorage


__all__ = ["MEGLocalStorage", "MEGRCloneStorage", "load_raw_file", "task_order"]


class MEGLocalStorage(LocalStorage):
    def __init__(self, root: str, tpt_name: TemplateName, task_name: str, scan_id: str):
        if task_name == "Rest":
            task_name = "Restin"
        super().__init__(root, [ANYTHING, "_MEG_", "scan_id", "-", "task_name",
                                ANYTHING, "megpreproc_rois-", "tpt_name", ".mat"],
                         **{"scan_id": scan_id, "task_name": task_name, "tpt_name": tpt_name})

    def get_all_by_scan(self):
        files = self.get_all()

        scans = {}
        for file in files:
            file_str = file.name
            found = re.findall('_[0-9]+-', file_str)
            if not len(found) == 1:
                print(f"WARNING: Cannot find unique scan id in {file_str}")
                continue
            scan_id = found[0][1:-1]

            found = re.findall('[0-9]+_MEG', file_str)
            if not len(found) == 1:
                print(f"WARNING: Cannot find unique subject id in {file_str}")
                continue
            subj_id = found[0].replace("_MEG", "")

            if scan_id not in scans:
                scans[scan_id] = []
            scans[scan_id].append((subj_id, file))

        return scans


class MEGRCloneStorage(RCloneStorage):
    def __init__(self, remote: str, root: str, task_name: str, scan_id: str):
        if task_name == "Rest":
            task_name = "Restin"
        super().__init__(
            remote, root, [ANYTHING, "_MEG_", "scan_id", "-", "task_name", ANYTHING, "megpreproc_voxels.mat"],
            **{"scan_id": scan_id, "task_name": task_name})

    def get_all_by_scan(self):
        files = self.get_all()

        scans = {}
        for file in files:
            file_str = file.name
            found = re.findall('_[0-9]+-', file_str)
            if not len(found) == 1:
                print(f"WARNING: Cannot find unique scan id in {file_str}")
                continue
            scan_id = found[0][1:-1]

            found = re.findall('[0-9]+_MEG', file_str)
            if not len(found) == 1:
                print(f"WARNING: Cannot find unique subject id in {file_str}")
                continue
            subj_id = found[0].replace("_MEG", "")

            if scan_id not in scans:
                scans[scan_id] = []
            scans[scan_id].append((subj_id, file))

        return scans


def load_raw_file(file: StorageFile):
    raw_data = loadmat(file.loadable_path)["data"]
    fs = raw_data["fsample"].item().item()
    data = np.concatenate(raw_data["trial"].item()[0], axis=1)

    return data, fs




import numpy as np
from joblib import Parallel, delayed
from scipy import signal
from statsmodels.tsa import stattools
from neuro_helper.statistics import fir_filter
from scipy.signal import hilbert


def _calc_parallel(func, data, n_job, threading, **kwargs):
    if n_job is None:
        n_job = data.shape[0]
    elif n_job == 0:
        raise ValueError("n_job cannot be zero")
    elif n_job > data.shape[0]:
        raise ValueError(f"n_job cannot be larger than the number of rows in data")
    backend = 'threading' if threading else None
    return np.asarray(Parallel(n_jobs=n_job, backend=backend)(delayed(func)(ts, **kwargs) for ts in data))


def lempel_ziv_complexity(sequence):
    sub_strings = set()

    ind = 0
    inc = 1
    while True:
        if ind + inc > len(sequence):
            break
        sub_str = sequence[ind: ind + inc]
        if sub_str in sub_strings:
            inc += 1
        else:
            sub_strings.add(sub_str)
            ind += inc
            inc = 1
    return len(sub_strings)


def calc_a_acf(ts, n_lag=None, fast=True):
    if not n_lag:
        n_lag = len(ts)
    return stattools.acf(ts, nlags=n_lag, qstat=False, alpha=None, fft=fast)


def calc_acf(data, n_lag=None, fast=True, n_job=None, threading=False):
    return _calc_parallel(calc_a_acf, data, n_job, threading, n_lag=n_lag, fast=fast)


def calc_a_acw(ts, n_lag=None, fast=True, is_acf=False):
    acf = ts if is_acf else calc_a_acf(ts, n_lag, fast)
    return np.argmax(acf < 0.5)


def calc_acw(data, n_lag=None, fast=True, is_acf=False, n_job=None, threading=False):
    return _calc_parallel(calc_a_acw, data, n_job, threading, n_lag=n_lag, fast=fast, is_acf=is_acf)


def calc_a_acz(ts, n_lag=None, fast=True, is_acf=False):
    acf = ts if is_acf else calc_a_acf(ts, n_lag, fast)
    return np.argmax(acf <= 0)


def calc_acz(data, n_lag=None, fast=True, is_acf=False, n_job=None, threading=False):
    return _calc_parallel(calc_a_acz, data, n_job, threading, n_lag=n_lag, fast=fast, is_acf=is_acf)


def calc_a_acmi(ts, which, n_lag=None, fast=True, is_acf=False):
    acf = ts if is_acf else calc_a_acf(ts, n_lag, fast)
    return signal.argrelextrema(acf, np.less)[0][which - 1]


def calc_acmi(data, which, n_lag=None, fast=True, is_acf=False, n_job=None, threading=False):
    return _calc_parallel(calc_a_acmi, data, n_job, threading, which=which, n_lag=n_lag, fast=fast, is_acf=is_acf)


def calc_a_acf_peaks(ts, cut_lag, is_acf=False):
    acf = ts if is_acf else calc_a_acf(ts)
    idx = np.concatenate((signal.argrelextrema(acf, np.less)[0], signal.argrelextrema(acf, np.greater)[0]))
    return np.sort(idx[idx < cut_lag])


def calc_acf_peaks(data, cut_lag, is_acf=False, n_job=None, threading=False):
    return _calc_parallel(calc_a_acf_peaks, data, n_job, threading, cut_lag=cut_lag, is_acf=is_acf)


def calc_lzc_norm_factor(ts):
    """
    The default way of calculating LZC normalization factor for a time series
    :param ts: a time series
    :return: normalization factor
    """
    return len(ts) / np.log2(len(ts))


def calc_a_lzc(ts, norm_factor=None):
    """
    Calculates lempel-ziv complexity of a single time series.
    :param ts: a time-series: nx1
    :param norm_factor: the normalization factor. If none, the default value will be calculated: n / ln(n)
    :return: the lempel-ziv complexity
    """
    if not norm_factor:
        norm_factor = calc_lzc_norm_factor(ts)
    bin_ts = np.char.mod('%i', ts >= np.median(ts))
    return lempel_ziv_complexity("".join(bin_ts)) / norm_factor


def calc_lzc(data, norm_factor=None, n_job=None, threading=False):
    return _calc_parallel(calc_a_lzc, data, n_job, threading, norm_factor=norm_factor)


def calc_mf(freq, psd):
    """
    Calculates median frequency of different power spectral densities.
    n is number of PSDs and m in number of frequencies.
    :param freq: the mx1 numpy array of frequencies.
    :param psd: the nxm numpy matrix of power.
    :return: nx1 array of median frequencies.
    """
    cum_sum = np.cumsum(psd, axis=1)
    return freq[np.argmax(cum_sum >= cum_sum[:, -1].reshape(-1, 1) / 2, axis=1)]


def calc_a_ple(a_psd, freq, is_log=False):
    if not is_log:
        freq = np.log(freq)
        a_psd = np.log(a_psd)
    return np.polyfit(freq, a_psd, deg=1)


def calc_ple(psd, freq, is_log=False, n_job=None, threading=False):
    return _calc_parallel(calc_a_ple, psd, n_job, threading, freq=freq, is_log=is_log)


def calc_peak(data, fs, low, high, apply_median_filter, n_job=None):
    data, freq_l, freq_h = fir_filter(data, fs, max_freq_low=low, min_freq_high=high, pass_type="bp")
    n_wave, n_sample = data.shape
    hilb_data = hilbert(data, n_sample, axis=0)
    phase = np.angle(hilb_data)
    instant_freq = fs * np.diff(np.unwrap(phase)) / (2 * np.pi)

    if not apply_median_filter:
        return instant_freq

    # apply median filter to attenuate large "blips" in instantaneous frequency
    n_window = 10
    win_lengths = np.round(np.linspace(10, 400, n_window) * fs / (2 * 1000)).astype(int)

    def win_calc(wl, si):
        start, stop = max(si - wl, 0), min(si + wl, n_sample)
        return np.median(np.sort(instant_freq[:, start: stop], axis=1), axis=1)

    n_job = n_job if n_job else 10
    phased_med = np.zeros((n_window, n_sample, n_wave))
    for wi, wl in enumerate(win_lengths):
        phased_med[wi, :, :] = np.asarray(Parallel(n_jobs=n_job)(delayed(win_calc)(wl, si) for si in range(n_sample)))
        # for si in range(n_sample):
        #     start, stop = max(si - win_lengths[wi], 0), min(si + win_lengths[wi], n_sample)
        #     phased_med[wi, si, :] = np.median(np.sort(instant_freq[:, start: stop], axis=1), axis=1)

    sliding_instant_freq = np.mean(phased_med, axis=0).T
    return sliding_instant_freq


def calc_ratio_occurrence(data_a, data_b, decimals=0, axis=None):
    precision = 1 / 10 ** decimals
    data_r = np.round(data_a / data_b, decimals)
    possible_ratios = np.arange(data_r.min(), data_r.max() + precision, precision)
    normalization_coeff = data_r.size if axis is None else data_r.shape[axis]
    return np.asarray([
        np.sum(data_r == np.round(ratio, decimals), axis=axis) / normalization_coeff * 100
        for ri, ratio in enumerate(possible_ratios)
    ])



from enum import Enum
import os
import glob
from abc import ABC, abstractmethod
from typing import List
from neuro_helper.rclone import RClone
import tempfile

ANYTHING = "*"


class StorageFile:
    local_path: str

    @property
    def name(self):
        return os.path.basename(self.local_path)

    @property
    def loadable_path(self):
        return self.local_path

    def __init__(self, local_path: str):
        self.local_path = local_path

    def __str__(self):
        return self.local_path

    def __lt__(self, other):
        return self.local_path < other.local_path


class RCloneStorageFile(StorageFile):
    remote_path: str
    engine: RClone
    downloaded = False

    @property
    def loadable_path(self):
        if not self.downloaded:
            self.download()
        return super().loadable_path

    def __init__(self, engine: RClone, remote_path: str):
        self.remote_path = remote_path
        super().__init__(os.path.join(tempfile.mkdtemp(), os.path.basename(remote_path)))
        self.engine = engine

    def download(self):
        if self.downloaded:
            return
        result = self.engine.copy(self.remote_path, self.local_path)
        self.downloaded = result["code"] == 0


class Storage(ABC):
    root: str
    names_pattern: str

    def __init__(self, root: str, patterns, **variables):
        self.root = root
        self.names_pattern = Storage.create_pattern(*patterns, **variables)

    @staticmethod
    def create_pattern(*patterns, **variables):
        if patterns is None:
            patterns = []
        if variables is None:
            variables = {}
        combined = ""
        for p in patterns:
            combined += str(p) if p not in variables else str(variables[p])
        return combined

    @abstractmethod
    def get_all(self, sort=True) -> List[StorageFile]:
        pass


class LocalStorage(Storage):
    def get_all(self, sort=True) -> List[StorageFile]:
        pattern = os.path.normpath(self.root + os.sep + self.names_pattern)
        files = list(map(lambda x: StorageFile(x), glob.glob(pattern)))
        if sort:
            files.sort()
        return files


class RCloneStorage(Storage):
    def __init__(self, remote: str, root: str, patterns, **variables):
        super().__init__(f"{remote}:{root}", patterns, **variables)
        self.engine = RClone()

    def get_all(self, sort=True) -> List[StorageFile]:
        files = list(map(lambda x: RCloneStorageFile(self.engine, os.path.join(self.root, x)),
                         filter(lambda x: x, self.engine.lsf(self.root, flags=["--include", self.names_pattern])["out"]
                                .decode("utf-8").split("\n"))))
        if sort:
            files.sort()
        return files



"""
A Python wrapper for rclone.
"""
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=W0102,W0703,C0103

import logging
import subprocess
import tempfile


__all__ = ["RClone"]


class RClone:
    """
    Wrapper class for rclone.
    """

    def __init__(self):
        self.log = logging.getLogger("RClone")

    def _execute(self, command_with_args):
        """
        Execute the given `command_with_args` using Popen

        Args:
            - command_with_args (list) : An array with the command to execute,
                                         and its arguments. Each argument is given
                                         as a new element in the list.
        """
        self.log.debug("Invoking : %s", " ".join(command_with_args))
        try:
            with subprocess.Popen(
                    command_with_args,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE) as proc:
                (out, err) = proc.communicate()

                #out = proc.stdout.read()
                #err = proc.stderr.read()

                self.log.debug(out)
                if err:
                    self.log.warning(err.decode("utf-8").replace("\\n", "\n"))

                return {
                    "code": proc.returncode,
                    "out": out,
                    "error": err
                }
        except FileNotFoundError as not_found_e:
            self.log.error("Executable not found. %s", not_found_e)
            return {
                "code": -20,
                "error": not_found_e
            }
        except Exception as generic_e:
            self.log.exception("Error running command. Reason: %s", generic_e)
            return {
                "code": -30,
                "error": generic_e
            }

    def run_cmd(self, command, extra_args=[]):
        """
        Execute rclone command

        Args:
            - command (string): the rclone command to execute.
            - extra_args (list): extra arguments to be passed to the rclone command
        """
        command_with_args = ["rclone", command]
        command_with_args += extra_args
        command_result = self._execute(command_with_args)
        return command_result

    def copy(self, source, dest, flags=[]):
        """
        Executes: rclone copy source:path dest:path [flags]

        Args:
        - source (string): A string "source:path"
        - dest (string): A string "dest:path"
        - flags (list): Extra flags as per `rclone copy --help` flags.
        """
        return self.run_cmd(command="copy", extra_args=[source] + [dest] + flags)

    def sync(self, source, dest, flags=[]):
        """
        Executes: rclone sync source:path dest:path [flags]

        Args:
        - source (string): A string "source:path"
        - dest (string): A string "dest:path"
        - flags (list): Extra flags as per `rclone sync --help` flags.
        """
        return self.run_cmd(command="sync", extra_args=[source] + [dest] + flags)

    def ls(self, dest, flags=[]):
        """
        Executes: rclone ls remote:path [flags]

        Args:
        - dest (string): A string "remote:path" representing the location to list.
        """
        return self.run_cmd(command="ls", extra_args=[dest] + flags)

    def lsf(self, dest, flags=[]):
        """
        Executes: rclone ls remote:path [flags]

        Args:
        - dest (string): A string "remote:path" representing the location to list.
        """
        return self.run_cmd(command="lsf", extra_args=[dest] + flags)

    def delete(self, dest, flags=[]):
        """
        Executes: rclone delete remote:path

        Args:
        - dest (string): A string "remote:path" representing the location to delete.
        """
        return self.run_cmd(command="delete", extra_args=[dest] + flags)